/****************************************************************************
** Copyright (C) 2020 MikroElektronika d.o.o.
** Contact: https://www.mikroe.com/contact
**
** Permission is hereby granted, free of charge, to any person obtaining a copy
** of this software and associated documentation files (the "Software"), to deal
** in the Software without restriction, including without limitation the rights
** to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
** copies of the Software, and to permit persons to whom the Software is
** furnished to do so, subject to the following conditions:
** The above copyright notice and this permission notice shall be
** included in all copies or substantial portions of the Software.
**
** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
** EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
** OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
** IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
** DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
** OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
**  USE OR OTHER DEALINGS IN THE SOFTWARE.
****************************************************************************/

/*!
 * @file c6dofimu23.c
 * @brief 6DOF IMU 23 Click Driver.
 */

#include "c6dofimu23.h"

/**
 * @brief Dummy data.
 * @details Definition of dummy data.
 */
#define DUMMY  0x00

void c6dofimu23_cfg_setup ( c6dofimu23_cfg_t *cfg ) 
{
    cfg->sck  = HAL_PIN_NC;
    cfg->miso = HAL_PIN_NC;
    cfg->mosi = HAL_PIN_NC;
    cfg->cs   = HAL_PIN_NC;
    cfg->rst  = HAL_PIN_NC;
    cfg->odr  = HAL_PIN_NC;

    cfg->spi_speed   = 100000;
    cfg->spi_mode    = SPI_MASTER_MODE_0;
    cfg->cs_polarity = SPI_MASTER_CHIP_SELECT_POLARITY_ACTIVE_LOW;
}

err_t c6dofimu23_init ( c6dofimu23_t *ctx, c6dofimu23_cfg_t *cfg ) 
{
    spi_master_config_t spi_cfg;

    spi_master_configure_default( &spi_cfg );

    spi_cfg.sck  = cfg->sck;
    spi_cfg.miso = cfg->miso;
    spi_cfg.mosi = cfg->mosi;

    ctx->chip_select = cfg->cs;

    if ( SPI_MASTER_ERROR == spi_master_open( &ctx->spi, &spi_cfg ) ) 
    {
        return SPI_MASTER_ERROR;
    }

    if ( SPI_MASTER_ERROR == spi_master_set_default_write_data( &ctx->spi, DUMMY ) ) 
    {
        return SPI_MASTER_ERROR;
    }

    if ( SPI_MASTER_ERROR == spi_master_set_mode( &ctx->spi, cfg->spi_mode ) ) 
    {
        return SPI_MASTER_ERROR;
    }

    if ( SPI_MASTER_ERROR == spi_master_set_speed( &ctx->spi, cfg->spi_speed ) ) 
    {
        return SPI_MASTER_ERROR;
    }

    spi_master_set_chip_select_polarity( cfg->cs_polarity );
    spi_master_deselect_device( ctx->chip_select );

    digital_out_init( &ctx->rst, cfg->rst );

    digital_in_init( &ctx->odr, cfg->odr );

    ctx->acc_fs = C6DOFIMU23_ACCEL_FS_DEFAULT;
    ctx->gyro_fs = C6DOFIMU23_GYRO_FS_DEFAULT;

    return SPI_MASTER_SUCCESS;
}

void c6dofimu23_device_enable ( c6dofimu23_t *ctx ) 
{
    digital_out_high( &ctx->rst );
}

uint8_t c6dofimu23_get_odr ( c6dofimu23_t *ctx ) 
{
    return digital_in_read( &ctx->odr );
}

err_t c6dofimu23_generic_write ( c6dofimu23_t *ctx, uint8_t offset, uint8_t *data_in, uint8_t len ) 
{
    spi_master_select_device( ctx->chip_select );
    err_t error_flag = spi_master_write( &ctx->spi, &offset, 1 );
    error_flag |= spi_master_write( &ctx->spi, data_in, len );
    spi_master_deselect_device( ctx->chip_select );
    return error_flag;
}

err_t c6dofimu23_generic_read ( c6dofimu23_t *ctx, uint8_t offset, uint8_t *data_out, uint8_t len ) 
{
    spi_master_select_device( ctx->chip_select );
    err_t error_flag = spi_master_write_then_read( &ctx->spi, &offset, 1, data_out, len );
    spi_master_deselect_device( ctx->chip_select );
    return error_flag;
}

err_t c6dofimu23_write_prtcl ( c6dofimu23_t *ctx, uint8_t *data_in ) 
{
    spi_master_select_device( ctx->chip_select );
    err_t err_flag = spi_master_write( &ctx->spi, data_in, 4 );
    spi_master_deselect_device( ctx->chip_select );
    return err_flag;
}

err_t c6dofimu23_read_prtcl ( c6dofimu23_t *ctx, uint8_t *data_in, uint16_t *data_out ) 
{
    uint8_t data_buf = DUMMY;
    spi_master_select_device( ctx->chip_select );
    err_t err_flag = spi_master_set_default_write_data( &ctx->spi, data_in[ 0 ] );
    err_flag |= spi_master_read( &ctx->spi, &data_buf, 1 );
    err_flag |= spi_master_set_default_write_data( &ctx->spi, data_in[ 1 ] );
    err_flag |= spi_master_read( &ctx->spi, &data_buf, 1 );
    *data_out = data_buf;
    *data_out <<= 8;
    err_flag |= spi_master_set_default_write_data( &ctx->spi, data_in[ 2 ] );
    err_flag |= spi_master_read( &ctx->spi, &data_buf, 1 );
    *data_out |= data_buf;
    err_flag |= spi_master_read( &ctx->spi, &data_buf, 1 );
    spi_master_deselect_device( ctx->chip_select );
    return err_flag;
}

err_t c6dofimu23_cmd_write ( c6dofimu23_t *ctx, uint32_t cmd ) 
{
    uint8_t data_buf[ 4 ] = { 0 };
    data_buf[ 0 ] = ( uint8_t ) ( cmd >> 24 );
    data_buf[ 1 ] = ( uint8_t ) ( cmd >> 16 );
    data_buf[ 2 ] = ( uint8_t ) ( cmd >> 8 );
    data_buf[ 3 ] = ( uint8_t )   cmd;
    return c6dofimu23_write_prtcl( ctx, data_buf );
}

err_t c6dofimu23_cmd_read ( c6dofimu23_t *ctx, uint32_t cmd, uint16_t *data_out ) 
{
    uint8_t data_buf[ 4 ] = { 0 };
    data_buf[ 0 ] = ( uint8_t ) ( cmd >> 24 );
    data_buf[ 1 ] = ( uint8_t ) ( cmd >> 16 );
    data_buf[ 2 ] = ( uint8_t ) ( cmd >> 8 );
    data_buf[ 3 ] = ( uint8_t )   cmd;
    return c6dofimu23_read_prtcl( ctx, data_buf, data_out );
}

err_t c6dofimu23_get_accel_data ( c6dofimu23_t *ctx, c6dofimu23_axis_t *acc_axis ) 
{
    uint16_t tmp_prtcl_buf = DUMMY;
    int16_t x_pos = DUMMY, y_pos = DUMMY, z_pos = DUMMY, x_neg = DUMMY, y_neg = DUMMY, z_neg = DUMMY;

    // Send SPI command to set Bank0
    err_t err_flag = c6dofimu23_cmd_write( ctx, C6DOFIMU23_CMD_SET_BANK_0 );

    // Send SPI command for DC self-test positive accel data
    err_flag |= c6dofimu23_cmd_write( ctx, C6DOFIMU23_CMD_ACCEL_DC_TRIGGER_POS );
    Delay_10ms( );
    Delay_10ms( );

    // Send SPI command to address accel X data
    err_flag |= c6dofimu23_cmd_write( ctx, C6DOFIMU23_CMD_ACCEL_X_DATA );

    // Send SPI command to address accel Y data and read positive accel X data
    err_flag |= c6dofimu23_cmd_read( ctx, C6DOFIMU23_CMD_ACCEL_Y_DATA, &tmp_prtcl_buf );
    x_pos = ( int16_t ) tmp_prtcl_buf;

    // Send SPI command to address accel Z data and read positive accel Y data
    err_flag |= c6dofimu23_cmd_read( ctx, C6DOFIMU23_CMD_ACCEL_Z_DATA, &tmp_prtcl_buf );
    y_pos = ( int16_t ) tmp_prtcl_buf;

    // Send SPI dummy command and read positive accel Z data
    err_flag |= c6dofimu23_cmd_read( ctx, C6DOFIMU23_CMD_DUMMY, &tmp_prtcl_buf );
    z_pos = ( int16_t ) tmp_prtcl_buf;

    // Send SPI command for DC self-test negative accel data
    err_flag |= c6dofimu23_cmd_write( ctx, C6DOFIMU23_CMD_ACCEL_DC_TRIGGER_NEG );
    Delay_10ms( );
    Delay_10ms( );

    // Send SPI command to address accel X data
    err_flag |= c6dofimu23_cmd_write( ctx, C6DOFIMU23_CMD_ACCEL_X_DATA );

    // Send SPI command to address accel Y data and read positive accel X data
    err_flag |= c6dofimu23_cmd_read( ctx, C6DOFIMU23_CMD_ACCEL_Y_DATA, &tmp_prtcl_buf );
    x_neg = ( int16_t ) tmp_prtcl_buf;

    // Send SPI command to address accel Z data and read positive accel Y data
    err_flag |= c6dofimu23_cmd_read( ctx, C6DOFIMU23_CMD_ACCEL_Z_DATA, &tmp_prtcl_buf );
    y_neg = ( int16_t ) tmp_prtcl_buf;
    
    // Send SPI dummy command and read positive accel Z data
    err_flag |= c6dofimu23_cmd_read( ctx, C6DOFIMU23_CMD_DUMMY, &tmp_prtcl_buf );
    z_neg = ( int16_t ) tmp_prtcl_buf;

    acc_axis->x = ( float ) x_pos;
    acc_axis->x -= ( float ) x_neg;
    acc_axis->x *= ctx->acc_fs;
    acc_axis->x /= C6DOFIMU23_DATA_RESOLUTION;
    
    acc_axis->y = ( float ) y_pos;
    acc_axis->y -= ( float ) y_neg;
    acc_axis->y *= ctx->acc_fs;
    acc_axis->y /= C6DOFIMU23_DATA_RESOLUTION;

    acc_axis->z = ( float ) z_pos;
    acc_axis->z -= ( float ) z_neg;
    acc_axis->z *= ctx->acc_fs;
    acc_axis->z /= C6DOFIMU23_DATA_RESOLUTION;

    // Send SPI command for DC self-test reset 
    err_flag |= c6dofimu23_cmd_write( ctx, C6DOFIMU23_CMD_DC_TRIGGER_RST );
    Delay_10ms( );
    Delay_10ms( );

    return err_flag;
}

err_t c6dofimu23_get_gyro_data ( c6dofimu23_t *ctx, c6dofimu23_axis_t *gyro_axis ) 
{
    uint16_t tmp_prtcl_buf = DUMMY;
    int16_t x_pos = DUMMY, y_pos = DUMMY, z_pos = DUMMY, x_neg = DUMMY, y_neg = DUMMY, z_neg = DUMMY;

    // Send SPI command to set Bank0
    err_t err_flag = c6dofimu23_cmd_write( ctx, C6DOFIMU23_CMD_SET_BANK_0 );

    // Send SPI command for DC self-test positive gyro data
    err_flag |= c6dofimu23_cmd_write( ctx, C6DOFIMU23_CMD_GYRO_DC_TRIGGER_POS );
    Delay_10ms( );
    Delay_10ms( );

    // Send SPI command to address gyro X data
    err_flag |= c6dofimu23_cmd_write( ctx, C6DOFIMU23_CMD_GYRO_X_DATA );

    // Send SPI command to address gyro Y data and read positive gyro X data
    err_flag |= c6dofimu23_cmd_read( ctx, C6DOFIMU23_CMD_GYRO_Y_DATA, &tmp_prtcl_buf );
    x_pos = ( int16_t ) tmp_prtcl_buf;

    // Send SPI command to address gyro Z data and read positive gyro Y data
    err_flag |= c6dofimu23_cmd_read( ctx, C6DOFIMU23_CMD_GYRO_Z_DATA, &tmp_prtcl_buf );
    y_pos = ( int16_t ) tmp_prtcl_buf;

    // Send SPI dummy command and read positive gyro Z data
    err_flag |= c6dofimu23_cmd_read( ctx, C6DOFIMU23_CMD_DUMMY, &tmp_prtcl_buf );
    z_pos = ( int16_t ) tmp_prtcl_buf;

    // Send SPI command for DC self-test negative gyro data
    err_flag |= c6dofimu23_cmd_write( ctx, C6DOFIMU23_CMD_GYRO_DC_TRIGGER_NEG );
    Delay_10ms( );
    Delay_10ms( );

    // Send SPI command to address gyro X data
    err_flag |= c6dofimu23_cmd_write( ctx, C6DOFIMU23_CMD_GYRO_X_DATA );

    // Send SPI command to address gyro Y data and read negative gyro X data
    err_flag |= c6dofimu23_cmd_read( ctx, C6DOFIMU23_CMD_GYRO_Y_DATA, &tmp_prtcl_buf );
    x_neg = ( int16_t ) tmp_prtcl_buf;

    // Send SPI command to address gyro Z data and read negative gyro Y data
    err_flag |= c6dofimu23_cmd_read( ctx, C6DOFIMU23_CMD_GYRO_Z_DATA, &tmp_prtcl_buf );
    y_neg = ( int16_t ) tmp_prtcl_buf;

    // Send SPI dummy command and read negative gyro Z data
    err_flag |= c6dofimu23_cmd_read( ctx, C6DOFIMU23_CMD_DUMMY, &tmp_prtcl_buf );
    z_neg = ( int16_t ) tmp_prtcl_buf;

    gyro_axis->x = ( int16_t ) x_pos;
    gyro_axis->x -= ( float ) x_neg;
    gyro_axis->x *= ctx->gyro_fs;
    gyro_axis->x /= C6DOFIMU23_DATA_RESOLUTION;

    gyro_axis->y = ( int16_t ) y_pos;
    gyro_axis->y -= ( float ) y_neg;
    gyro_axis->y *= ctx->gyro_fs;
    gyro_axis->y /= C6DOFIMU23_DATA_RESOLUTION;

    gyro_axis->z = ( int16_t ) z_pos;
    gyro_axis->z -= ( float ) z_neg;
    gyro_axis->z *= ctx->gyro_fs;
    gyro_axis->z /= C6DOFIMU23_DATA_RESOLUTION;

    // Send SPI command for DC self-test reset 
    err_flag |= c6dofimu23_cmd_write( ctx, C6DOFIMU23_CMD_DC_TRIGGER_RST );
    Delay_10ms( );
    Delay_10ms( );

    return err_flag;
}

// ------------------------------------------------------------------------- END
